import numpy as np
from numpy.fft import fft2, ifft2, fftshift, ifftshift
import skimage.io as io
import matplotlib.pyplot as plt
from skimage.metrics import peak_signal_noise_ratio as compute_psnr
from skimage.metrics import structural_similarity as compute_ssim
from pypher.pypher import psf2otf
from skimage.filters import gaussian
from scipy.ndimage import median_filter
import bm3d

from utils import *


img = io.imread('./Si/Si_raw.png').astype(float)/255
#noisy = io.imread('./Si/Si_noisy.png').astype(float)/255
#img = io.imread('./flatfield/noise analysis_0001.png')

def noise_analysis(img):
    plt.hist(img.reshape(-1,), bins = 100)
    plt.ylabel('Counts')
    plt.xlabel('Pixel Value');
    plt.savefig('noise_dose=1.png')


'''
simulate nosiy TEM image
@ img: ground truth image
@ sigma: noise level
'''
def simulate(img, sigma):
    c = fspecial_gaussian_2d((10, 10), 0.8)
    cFT = psf2otf(c, img.shape)
    blurred = np.abs(ifft2((np.multiply(fft2(img), cFT))))     # blurred with Gaussain kernel
    corrupted = blur + sigma * np.random.randn(*blurred.shape) # add noise
    subsampled = corrupted[::2, ::2]
    return subsampled

'''
gaussain filtering
'''
def gaussian_filter(img, sigma):
    return gaussian(img, sigma)

'''
median filtering
'''
def median_filter(img, filtSize):
    return median_filter(img, filtSize)

'''
weiner filtering
'''
def weiner_filter(img, sigma):
    c = fspecial_gaussian_2d((10, 10), sigma)
    cFT = psf2otf(c, img.shape)
    inverse_SNR = sigma / np.mean(img)
    damp_kernel = np.power(cFT, 2) 
    damp_factor = np.divide(damp_kernel, damp_kernel + inverse_SNR)
    weiner_filtered = np.abs(ifft2(np.multiply(np.divide(fft2(img), cFT), damp_factor)))
    return weiner_filtered

'''
admm+tv
'''
def admmtv(img):
    return admm_tv_iso(img)

'''
Non-local means filtering
'''
sigma = 0.8
nlmSigma = 0.05 
searchWindowRadius = 11
averageFilterRadius = 7
def nlm(img):
    return nonlocalmeans(noisy, searchWindowRadius, averageFilterRadius, sigma, nlmSigma)

'''
local adaptive non-local means
'''
sigma = 0.8
searchWindowRadius = 13
averageFilterRadius = 7
def lanlm(img):
    return localadaptivenonlocalmeans(noisy, searchWindowRadius, averageFilterRadius, sigma)

'''
bm3d
'''
def bm3d_filter(img, signa_psd):
    return bm3d.bm3d(noisy, sigma_psd = 0.1, stage_arg = bm3d.BM3DStages.ALL_STAGES)


