
import numpy as np
from pypher.pypher import psf2otf
from numpy.fft import fft2, ifft2, fftshift, ifftshift
from skimage.filters import gaussian
from tqdm import tqdm


def fspecial_gaussian_2d(size, sigma):
    kernel = np.zeros(tuple(size))
    kernel[size[0]//2, size[1]//2] = 1
    kernel = gaussian(kernel, sigma)
    return kernel/np.sum(kernel)


def iso_soft_threshold(v, k):
    ret = np.zeros_like(v)
    thr = np.maximum(1 - np.divide(k, np.linalg.norm(v, axis=2)),0)
    for i in range(v.shape[-1]):
        ret[:,:,i] = np.multiply(thr, v[:,:,i])
    return ret

def admm_tv_iso(img):
	rou = 10
    # Convolution kernel for Dx and Dy
    dx = np.array([[-1., 1.]])
    dy = np.array([[-1.], [1.]])
    # Precompute OTFs
    cFT = psf2otf(c, noisy.shape)     # F(c)
    cTFT = np.conj(cFT)             # F(c)*
    dxFT = psf2otf(dx, noisy.shape)   # F(d_x)
    dxTFT = np.conj(dxFT)           # F(d_x)*
    dyFT = psf2otf(dy, noisy.shape)   # F(d_y)
    dyTFT = np.conj(dyFT)           # F(d_y)*
    iters = list(range(1,101))
    simga = 0.2
    lam = 0.001
    # Add noise to blurred image
    b = np.power(noisy, 1/2.2)
    bFT = fft2(b)
    kappa = lam / rou
    ### Deconvolution using ADMM here ###
    x = np.zeros_like(noisy)
    z = np.zeros((noisy.shape[0], noisy.shape[1], 2))
    u = np.zeros_like(z)
    Dx = np.zeros_like(z)
    Dx[:,:,0] = ifft2(np.multiply(fft2(x), dxFT)).real
    Dx[:,:,1] = ifft2(np.multiply(fft2(x), dyFT)).real
    v = Dx + u
    x_denom = np.multiply(cTFT, cFT) + rou * (np.multiply(dxTFT, dxFT) + np.multiply(dyTFT, dyFT))
    for k in tqdm(range(100)):
        v = z - u
        #update x
        v1FT = fft2(v[:,:,0])
        v2FT = fft2(v[:,:,1])
        x_numer = np.multiply(cTFT, bFT) + rou * (np.multiply(dxTFT, v1FT) + \
                np.multiply(dyTFT, v2FT))
        x = ifft2(np.divide(x_numer, x_denom)).real
        #update z
        Dx[:,:,0] = ifft2(np.multiply(fft2(x), dxFT)).real
        Dx[:,:,1] = ifft2(np.multiply(fft2(x), dyFT)).real
        v = Dx + u
        #z = aniso_soft_threshold(v, kappa)
        z = iso_soft_threshold(v, kappa)
        #update u
        u = u + Dx - z
    return x


def inbounds(img, y, x):
    return 0 <= y and y < img.shape[0] and \
           0 <= x and x < img.shape[1]

def comparePatches(patch1, patch2, kernel, sigma):
    return np.exp(-np.sum(kernel*(patch1 - patch2) ** 2)/(2*sigma**2))

def comparePatchesLA(patch1, patch2, kernel, nlmsigma, beta):
    return np.exp(-np.sum(kernel*(patch1 - patch2) ** 2)/((beta * nlmsigma)**2))

def getnlmsigma(local_var, img_var):
    nlmsigma = img_var
    if local_var < 0.7 * img_var:
        nlmsigma = 0.7 * img_var
    elif local_var > 1.2 * img_var:
        nlmsigma = 1.2 * img_var
    return nlmsigma

def nonlocalmeans(img, searchWindowRadius, averageFilterRadius, sigma, nlmSigma):
    # Initialize output to 0
    out = np.zeros_like(img)
    # Pad image to reduce boundary artifacts
    pad = max(averageFilterRadius, searchWindowRadius)
    imgPad = np.pad(img, pad, mode = "symmetric")

    # Smoothing kernel
    filtSize = (2*averageFilterRadius + 1, 2*averageFilterRadius + 1)
    kernel = fspecial_gaussian_2d(filtSize, sigma)
    #radius = searchWindowRadius-averageFilterRadius
    for y in range(img.shape[0]): #img[y, x]
        for x in range(img.shape[1]):
            centerPatch = imgPad[y+pad-averageFilterRadius:y+pad+averageFilterRadius+1,
                                 x+pad-averageFilterRadius:x+pad+averageFilterRadius+1]
            cur_pixel_nlm = 0
            weight_max = 0
            weight_sum = 0
            for j in range(y-searchWindowRadius, y+searchWindowRadius+1):
                for i in range(x-searchWindowRadius, x+searchWindowRadius+1):
                    if ((not (j == y and i == x)) and inbounds(img, j, i)): 
                        searchPatch = imgPad[j+pad-averageFilterRadius:j+pad+averageFilterRadius+1,
                                            i+pad-averageFilterRadius:i+pad+averageFilterRadius+1]
                        weight = comparePatches(centerPatch, searchPatch, kernel, nlmSigma)
                        if weight > weight_max: # udpate weight_max
                            weight_max = weight
                        cur_pixel_nlm  += img[j,i] * weight
                        weight_sum += weight
            cur_pixel_nlm  += weight_max * img[y,x] # add current pixel
            weight_sum += weight_max                  # add weight_max for current pixel
            out[y, x] = cur_pixel_nlm  / weight_sum
    return out

def localadaptivenonlocalmeans(img, searchWindowRadius, averageFilterRadius, sigma):
    # Initialize output to 0
    out = np.zeros_like(img)
    # Pad image to reduce boundary artifacts
    pad = max(averageFilterRadius, searchWindowRadius)
    imgPad = np.pad(img, pad, mode = "symmetric")
    img_var = np.var(img)
    beta = 5
    # Smoothing kernel
    filtSize = (2*averageFilterRadius + 1, 2*averageFilterRadius + 1)
    kernel = fspecial_gaussian_2d(filtSize, sigma)
    #radius = searchWindowRadius-averageFilterRadius
    for y in range(img.shape[0]): #img[y, x]
        for x in range(img.shape[1]):
            centerPatch = imgPad[y+pad-averageFilterRadius:y+pad+averageFilterRadius+1,
                                 x+pad-averageFilterRadius:x+pad+averageFilterRadius+1]
            cur_pixel_nlm = 0
            weight_max = 0
            weight_sum = 0
            for j in range(y-searchWindowRadius, y+searchWindowRadius+1):
                for i in range(x-searchWindowRadius, x+searchWindowRadius+1):
                    if ((not (j == y and i == x)) and inbounds(img, j, i)): 
                        searchPatch = imgPad[j+pad-averageFilterRadius:j+pad+averageFilterRadius+1,
                                            i+pad-averageFilterRadius:i+pad+averageFilterRadius+1]
                        local_var = np.var(searchPatch)
                        nlmsigma = getnlmsigma(local_var, img_var)
                        weight = comparePatchesLA(centerPatch, searchPatch, kernel, nlmsigma, beta)
                        if weight > weight_max: # udpate weight_max
                            weight_max = weight
                        cur_pixel_nlm  += img[j,i] * weight
                        weight_sum += weight
            cur_pixel_nlm  += weight_max * img[y,x] 
            weight_sum += weight_max               
            out[y, x] = cur_pixel_nlm  / weight_sum
    return out

def minmax(img):
	return  (img - np.min(img)) / (np.max(img) - np.min(img))
